@@ -1,3 +1,5 @@
+# pylint: disable=missing-docstring
+
 # Copyright 2017-2018 Nir Cohen
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -18,32 +20,32 @@ import uuid
 import json
 import socket
 import logging
-import datetime
+import logging.handlers
+from datetime import datetime
+
+# pytype: disable=import-error
+try:
+    # Python 2
+    import urllib2 as urllib
+except ImportError:
+    # Python 3
+    import urllib.request as urllib
 
 try:
     import colorama
     from colorama import Fore, Style
+
     COLOR_ENABLED = True
 except ImportError:
     COLOR_ENABLED = False
 
 try:
-    import click
+    import click  # pytype: disable=pyi-error
+
     CLI_ENABLED = True
 except ImportError:
     CLI_ENABLED = False
-
-try:
-    from logzio.handler import LogzioHandler
-    LOGZIO_INSTALLED = True
-except ImportError:
-    LOGZIO_INSTALLED = False
-
-try:
-    from cmreslogging.handlers import CMRESHandler
-    ELASTICSEARCH_INSTALLED = True
-except ImportError:
-    ELASTICSEARCH_INSTALLED = False
+# pytype: enable=import-error
 
 
 LEVEL_CONVERSION = {
@@ -53,7 +55,7 @@ LEVEL_CONVERSION = {
     'warn': logging.WARN,
     'error': logging.ERROR,
     'critical': logging.CRITICAL,
-    'event': logging.INFO
+    'event': logging.INFO,
 }
 
 
@@ -62,7 +64,6 @@ class JsonFormatter(logging.Formatter):
         self.pretty = pretty
 
     def format(self, record):
-        # TODO: Allow to use ujson or radpijson via config
         return json.dumps(record.msg, indent=4 if self.pretty else None)
 
 
@@ -74,7 +75,7 @@ class ConsoleFormatter(logging.Formatter):
         _simple = os.getenv('WRYTE_SIMPLE_CONSOLE')
 
         if _simple is not None:
-            self.simple = True if _simple == "true" else False
+            self.simple = _simple == "true"
         else:
             self.simple = simple
 
@@ -112,58 +113,47 @@ class ConsoleFormatter(logging.Formatter):
         Performance is also reduced by the amount of fields you have in your
         context (context i.e. k=v).
         """
-        # TODO: No need to copy here
         record = record.msg.copy()
 
-        # TODO: pop instead so that we don't need to pop after
+        # Not popping and deleting later as pop is marginally less performant
         name = record['name']
         timestamp = record['timestamp']
         level = record['level'] if record['type'] == 'log' else 'EVENT'
         message = record['message']
-
         # We no longer need them as part of the dict.
-        p = ('name', 'timestamp', 'level', 'message',
-             'type', 'hostname', 'pid')
-        for key in p:
-            # TODO: del instead of popping
-            record.pop(key)
+        drop_keys = ('level', 'type', 'hostname', 'pid', 'name', 'message', 'timestamp')
+        for key in drop_keys:
+            del record[key]
 
         if COLOR_ENABLED and self.color and not self.simple:
-            # TODO: Use string formatting instead
             level = str(self._get_level_color(level) + level + Style.RESET_ALL)
             timestamp = str(Fore.GREEN + timestamp + Style.RESET_ALL)
             name = str(Fore.MAGENTA + name + Style.RESET_ALL)
 
-        if self.simple:
-            msg = message
-        else:
-            # TODO: Use ' - '.join((timestamp, name, level, message))
-            msg = '{0} - {1} - {2} - {3}'.format(
-                timestamp, name, level, message)
+        msg = message if self.simple else ' - '.join((timestamp, name, level, message))
 
         if self.pretty:
-            # TODO: Find an alternative to concat here
-            # msg += ''.join("\n  %s=%s" % (k, v)
-            #                for (k, v) in record.items())
-            for key, value in record.items():
-                msg += '\n  {0}={1}'.format(key, value)
+            # https://codereview.stackexchange.com/questions/7953/flattening-a-dictionary-into-a-string
+            msg += ''.join("\n  %s=%s" % item for item in record.items())
         elif record:
-            # TODO: Allow to use ujson or radpijson
-            msg += '\n{0}'.format(json.dumps(record, indent=4))
+            msg += '\n{}'.format(json.dumps(record, indent=4))
 
         return msg
 
 
-class Wryte(object):
-    def __init__(self,
-                 name=None,
-                 hostname=None,
-                 level='info',
-                 pretty=None,
-                 bare=False,
-                 jsonify=False,
-                 color=True,
-                 simple=False):
+class Wryte:
+    def __init__(
+        self,
+        name=None,
+        hostname=None,
+        level='info',
+        pretty=None,
+        bare=False,
+        jsonify=False,
+        color=True,
+        simple=False,
+        enable_ec2=False,
+    ):
         """Instantiate a logger instance.
 
         Either a JSON or a Console handler will be added to the logger
@@ -184,40 +174,67 @@ class Wryte(object):
         self.color = color
         self.simple = simple
 
-        self._log = self._get_base(self.logger_name, hostname)
         self.logger = self._logger(self.logger_name)
+        self._log = self._get_base(self.logger_name, hostname, enable_ec2)
 
         if not bare:
             self._configure_handlers(level, jsonify)
 
     @staticmethod
     def _logger(name):
-        """Return a named logger instance.
-        """
+        """Return a named logger instance."""
         logger = logging.getLogger(name)
         return logger
 
-    @staticmethod
-    def _get_base(name, hostname):
+    def _get_base(self, name, hostname, enable_ec2=False):
         """Generate base fields for each log message.
 
         This is evaluated once when the logger's instance is instantiated.
         It is then later copied by each log message.
         """
-        # TODO: Document that these are only generated once.
-        return {
+
+        def fetch_ec2(attribute):
+            try:
+                return urllib.urlopen('http://169.254.169.254/latest/meta-data/{}'.format(attribute)).read().decode()
+            # Yuch. But shouldn't take a risk that any exception will raise
+            except Exception:  # pylint: disable=broad-except
+                return None
+
+        base = {
             'name': name,
             'hostname': hostname or socket.gethostname(),
             'pid': os.getpid(),
-            'type': 'log'
+            'type': 'log',
         }
 
+        if self._env('EC2_ENABLED') or enable_ec2:
+            # To test that ec2 data is actually attainable
+            instance_id = fetch_ec2('instance-id')
+            if instance_id:
+                base['ec2_instance_id'] = instance_id
+                base['ec2_instance_type'] = fetch_ec2('instance-type')
+                base['ec2_region'] = fetch_ec2('placement/availability-zone')
+                base['ec2_ipv4'] = fetch_ec2('local-ipv4')
+            else:
+                self.logger.error(
+                    'WRYTE EC2 env var set but EC2 metadata endpoint is '
+                    'unavailable or the data could not be retrieved.'
+                )
+
+        return base
+
     @staticmethod
     def _get_timestamp():
-        # TODO: Allow to use udatetime instead for faster evals
-        return datetime.datetime.now().isoformat()
+        # `now()` needs to compensate for timezones, and so it takes much
+        # more time to evaluate. `udatetime` doesn't help here and actually
+        # takes more time both on Python2 and Python3.
+        # This is by no means a reason to use utcnow,
+        # but since we should standardize the timestamp, it makes sense to do
+        # so anyway.
+        return datetime.utcnow().isoformat()
 
-    def _normalize_objects(self, objects):
+    @staticmethod
+    def _normalize_objects(objects):
         """Return a normalized dictionary for a list of key value like objects.
 
         This supports parsing dicts, json strings and key=value pairs.
@@ -228,19 +245,19 @@ class Wryte(object):
         A `bad_object_uuid` field will be added to the context if an object
         doesn't fit the supported formats.
         """
-        # TODO: Generate a consolidated dict instead of a list of objects
         consolidated = {}
 
         for obj in objects:
-            try:
+            # We if here instead of try-excepting since it's not obvious
+            # what the distribution between dict and json will be and if
+            # costs much less when the distribution is flat.
+            if isinstance(obj, dict):
                 consolidated.update(obj)
-            except ValueError:
+            else:
                 try:
                     consolidated.update(json.loads(obj))
-                # # TODO: Should be a JsonDecoderError
-                except Exception:  # NOQA
-                    consolidated.update(
-                        {'_bad_object_{0}'.format(str(uuid.uuid4())): obj})
+                except Exception:  # pylint: disable=broad-except
+                    consolidated['_bad_object_{}'.format(str(uuid.uuid4()))] = obj
         return consolidated
 
     def _enrich(self, message, level, objects, kwargs=None):
@@ -274,17 +291,9 @@ class Wryte(object):
             log.update(kwargs)
 
         # Appends default fields.
-        # This of course means that if any of these are provided
-        # within the chain, they will be overriden here.
-        log.update({
-            'message': message,
-            # TODO: declare `upper` method when instantiating class or
-            # simply remove it altogether.
-            'level': level.upper(),
-            # TODO: Maybe we don't need a method here and can directly
-            # call the datetime method? Perf-wise, that is..
-            'timestamp': self._get_timestamp()
-        })
+        log['message'] = message
+        log['level'] = level.upper()
+        log['timestamp'] = self._get_timestamp()
 
         return log
 
@@ -301,9 +310,8 @@ class Wryte(object):
         Setting the variable `WRYTE_HANDLERS_LOGZIO_TOKEN` means
         that it applies to all loggers.
         """
-        logger_env = os.getenv('WRYTE_{0}_{1}'.format(
-            self.logger_name.upper(), variable))
-        global_env = os.getenv('WRYTE_{0}'.format(variable))
+        logger_env = os.getenv('WRYTE_{}_{}'.format(self.logger_name.upper(), variable))
+        global_env = os.getenv('WRYTE_{}'.format(variable))
         return logger_env or global_env or default
 
     def _configure_handlers(self, level, jsonify=False):
@@ -317,18 +325,9 @@ class Wryte(object):
             else:
                 self.add_default_console_handler(level)
 
-        if self._env('HANDLERS_SYSLOG_ENABLED'):
-            self.add_syslog_handler()
-
-        if self._env('HANDLERS_LOGZIO_ENABLED'):
-            self.add_logzio_handler()
-
-        if self._env('HANDLERS_FILE_ENABLED'):
+        if self._env('HANDLERS_FILE_PATH'):
             self.add_file_handler()
 
-        if self._env('HANDLERS_ELASTICSEARCH_ENABLED'):
-            self.add_elasticsearch_handler()
-
     def _assert_level(self, level):
         levels = LEVEL_CONVERSION.keys()
 
@@ -337,11 +336,7 @@ class Wryte(object):
             return False
         return True
 
-    def add_handler(self,
-                    handler,
-                    name=None,
-                    formatter='json',
-                    level='info'):
+    def add_handler(self, handler, name=None, formatter='json', level='info'):
         """Add a handler to the logger instance and return its name.
 
         A `handler` can be any standard `logging` handler.
@@ -355,35 +350,34 @@ class Wryte(object):
         if self._assert_level(level):
             self.logger.setLevel(LEVEL_CONVERSION[level.lower()])
         else:
-            return
+            return ''
 
-        # TODO: Allow to ignore fields in json formatter
-        # TODO: Allow to remove field printing in console formatter
         if formatter == 'json':
             _formatter = JsonFormatter(self.pretty or False)
         elif formatter == 'console':
             if COLOR_ENABLED:
                 colorama.init(autoreset=True)
-            pretty = True if self.pretty in (None, True) else False
+            pretty = self.pretty in (None, True)
             _formatter = ConsoleFormatter(pretty, self.color, self.simple)
         else:
             _formatter = formatter
 
         handler.setFormatter(_formatter)
-        handler.set_name(name)
+        try:
+            handler.set_name(name)  # pytype: disable=attribute-error
+        except AttributeError:
+            handler.name = name
 
         self.logger.addHandler(handler)
 
         return name
 
     def list_handlers(self):
-        """Return a list of all handlers attached to a logger
-        """
+        """Return a list of all handlers attached to a logger"""
         return [handler.name for handler in self.logger.handlers]
 
     def remove_handler(self, name):
-        """Remove a handler by its name (set in `add_handler`)
-        """
+        """Remove a handler by its name (set in `add_handler`)"""
         for handler in self.logger.handlers:
             if handler.name == name:
                 self.logger.removeHandler(handler)
@@ -393,7 +387,8 @@ class Wryte(object):
             handler=logging.StreamHandler(sys.stdout),
             name='_json',
             formatter='json',
-            level=self._env('CONSOLE_LEVEL', level))
+            level=self._env('CONSOLE_LEVEL', level),
+        )
 
     def add_default_console_handler(self, level='debug'):
         name = '_console'
@@ -401,15 +396,12 @@ class Wryte(object):
         formatter = 'console'
         handler = logging.StreamHandler(sys.stdout)
 
-        return self.add_handler(
-            handler=handler,
-            name=name,
-            formatter=formatter,
-            level=level)
+        return self.add_handler(handler=handler, name=name, formatter=formatter, level=level)
 
-    def add_file_handler(self, **kwargs):
+    def add_file_handler(self):
         if not self._env('HANDLERS_FILE_PATH'):
-            self.logger.warn('File handler file path not set')
+            self.logger.warning('File handler file path not set')
+            return
 
         name = self._env('HANDLERS_FILE_NAME', default='file')
         level = self._env('HANDLERS_FILE_LEVEL', default='info')
@@ -417,129 +409,24 @@ class Wryte(object):
 
         if self._env('HANDLERS_FILE_ROTATE'):
             try:
-                max_bytes = int(
-                    self._env('HANDLERS_FILE_MAX_BYTES', default=13107200))
-                backup_count = int(
-                    self._env('HANDLERS_FILE_BACKUP_COUNT', default=7))
+                max_bytes = int(self._env('HANDLERS_FILE_MAX_BYTES', default=13107200))
+                backup_count = int(self._env('HANDLERS_FILE_BACKUP_COUNT', default=7))
             except ValueError:
-                self.logger.exception(
-                    'MAX_BYTES and BACKUP_COUNT must be integers')
+                self.logger.exception('MAX_BYTES and BACKUP_COUNT must be integers')
                 return
 
             handler = logging.handlers.RotatingFileHandler(
-                self._env('HANDLERS_FILE_PATH'),
-                maxBytes=max_bytes,
-                backupCount=backup_count)
+                self._env('HANDLERS_FILE_PATH'), maxBytes=max_bytes, backupCount=backup_count
+            )
         elif os.name == 'nt':
             handler = logging.FileHandler(self._env('HANDLERS_FILE_PATH'))
         else:
-            handler = logging.handlers.WatchedFileHandler(
-                self._env('HANDLERS_FILE_PATH'))
+            handler = logging.handlers.WatchedFileHandler(self._env('HANDLERS_FILE_PATH'))
 
-        self.add_handler(
-            handler=handler,
-            name=name,
-            formatter=formatter,
-            level=level)
-
-    def add_syslog_handler(self, **kwargs):
-        name = self._env('HANDLERS_SYSLOG_NAME', default='syslog')
-        level = self._env('HANDLERS_SYSLOG_LEVEL', default='info')
-        formatter = self._env('HANDLERS_SYSLOG_FORMATTER', default='json')
-
-        syslog_host = self._env('HANDLERS_SYSLOG_HOST',
-                                default='localhost:514')
-        syslog_host = syslog_host.split(':', 1)
-
-        if len(syslog_host) == 2:
-            # Syslog listener
-            host, port = syslog_host
-            address = (host, port)
-        else:
-            # Unix socket or otherwise
-            address = syslog_host
-
-        socket_type = self._env('HANDLERS_SYSLOG_SOCKET_TYPE', default='udp')
-        if socket_type not in ('tcp', 'udp'):
-            self.logger.warn(
-                'syslog handler socket type must be one of tcp/udp')
-
-        handler = logging.handlers.SysLogHandler(
-            address=address,
-            facility=self._env('HANDLERS_SYSLOG_FACILITY', default='LOG_USER'),
-            socktype=socket.SOCK_STREAM if socket_type == 'tcp'
-            else socket.SOCK_DGRAM)
-
-        self.add_handler(
-            handler=handler,
-            name=name,
-            formatter=formatter,
-            level=level)
-
-    def add_logzio_handler(self, **kwargs):
-        if LOGZIO_INSTALLED:
-            if not self._env('HANDLERS_LOGZIO_TOKEN'):
-                self.logger.warn('Logzio handler token not set')
-
-            name = self._env('HANDLERS_LOGZIO_NAME', default='logzio')
-            level = self._env('HANDLERS_LOGZIO_LEVEL', default='info')
-            formatter = self._env('HANDLERS_LOGZIO_FORMATTER', default='json')
-            handler = LogzioHandler(self._env('HANDLERS_LOGZIO_TOKEN'))
-
-            self.add_handler(
-                handler=handler,
-                name=name,
-                formatter=formatter,
-                level=level)
-        else:
-            self.logger.error(
-                'It seems that the logzio handler is not installed. '
-                'You can install it by running `pip install '
-                'wryte[logzio]`')
-
-    def add_elasticsearch_handler(self, **kwargs):
-        if ELASTICSEARCH_INSTALLED:
-            if not self._env('HANDLERS_ELASTICSEARCH_HOST'):
-                self.logger.warn('Elasticsearch handler host not set')
-
-            name = self._env('HANDLERS_ELASTICSEARCH_NAME',
-                             default='elasticsearch')
-            level = self._env('HANDLERS_ELASTICSEARCH_LEVEL', default='info')
-            formatter = self._env(
-                'HANDLER_ELASTICSEARCH_FORMATTER', default='json')
-
-            hosts = []
-            es_hosts = self._env('HANDLERS_ELASTICSEARCH_HOST')
-            es_hosts = es_hosts.split(',')
-            for es_host in es_hosts:
-                host, port = es_host.split(':', 1)
-                hosts.append({'host': host, 'port': port})
-
-            handler_args = {
-                'hosts': hosts,
-                'auth_type': CMRESHandler.AuthType.NO_AUTH
-            }
-
-            handler = CMRESHandler(**handler_args)
-
-            self.add_handler(
-                handler=handler,
-                name=name,
-                formatter=formatter,
-                level=level)
-        else:
-            self.logger.error(
-                'It seems that the elasticsearch handler is not installed. '
-                'You can install it by running `pip install '
-                'wryte[elasticsearch]`')
+        self.add_handler(handler=handler, name=name, formatter=formatter, level=level)
 
     def set_level(self, level):
-        """Set the current logger instance's level.
-        """
-        # TODO: Consider removing this check and letting the user
-        # take the hit incase they provide an unreasonable level.
-        # This would reduce overhead when using `set_level` in
-        # error messages under heavy load.
+        """Set the current logger instance's level."""
         if not self._assert_level(level):
             return
 
@@ -556,10 +443,12 @@ class Wryte(object):
             self._log.update(kwargs)
 
     def unbind(self, *keys):
-        """Unbind previously bound context.
-        """
-        # TODO: Support unbinding nested field in context
+        """Unbind previously bound context."""
         for key in keys:
+            # Perf-wise, we should try-except here since we expect
+            # that 99% of the time the keys will exist so it will be faster.
+            # Thing is, that unbinding shouldn't happen thousands of
+            # times a second, so we'll go for readability here.
             self._log.pop(key)
 
     def event(self, message, *objects, **kwargs):
@@ -569,9 +458,7 @@ class Wryte(object):
         explicitly passed in kwargs. Additionally, the `type` of the
         log will be `event`, instead of log, like in other cases.
         """
-        # TODO: Prefix cid key with underscore
         cid = kwargs.get('cid', str(uuid.uuid4()))
-        # TODO: Consider allowing to bind `cid` here.
         objects = objects + ({'type': 'event', 'cid': cid},)
         obj = self._enrich(message, 'info', objects, kwargs)
         self.logger.info(obj)
@@ -592,8 +479,8 @@ class Wryte(object):
         can practically pass weird logging levels.
         """
         obj = self._enrich(message, level, objects, kwargs)
-        if kwargs.get('_set_level'):
-            self.set_level(kwargs.get('_set_level'))
+        if '_set_level' in kwargs:
+            self.set_level(kwargs['_set_level'])
 
         if not self._assert_level(level):
             return
@@ -619,20 +506,14 @@ class Wryte(object):
         self.logger.warning(obj)
 
     def error(self, message, *objects, **kwargs):
-        if kwargs.get('_set_level'):
-            # TODO: Use subscriptiong instead
-            self.set_level(kwargs.get('_set_level'))
-            # TODO: Don't pop, this could be useful in the log
-            kwargs.pop('_set_level')
+        if '_set_level' in kwargs:
+            self.set_level(kwargs['_set_level'])
         obj = self._enrich(message, 'error', objects, kwargs)
         self.logger.error(obj)
 
     def critical(self, message, *objects, **kwargs):
-        if kwargs.get('_set_level'):
-            # TODO: Use subscriptiong instead
-            self.set_level(kwargs.get('_set_level'))
-            # TODO: Don't pop, this could be useful in the log
-            kwargs.pop('_set_level')
+        if '_set_level' in kwargs:
+            self.set_level(kwargs['_set_level'])
         obj = self._enrich(message, 'critical', objects, kwargs)
         self.logger.critical(obj)
 
@@ -641,54 +522,54 @@ class WryteError(Exception):
     pass
 
 
+def _split_kv(pair):
+    """Return dict for key=value."""
+    key_value = pair.split('=', 1)
+    return {key_value[0]: key_value[1]}
+
+
 if CLI_ENABLED:
-    CLICK_CONTEXT_SETTINGS = dict(
-        help_option_names=['-h', '--help'],
-        token_normalize_func=lambda param: param.lower())
+    CLICK_CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'], token_normalize_func=lambda param: param.lower())
 
     @click.command(context_settings=CLICK_CONTEXT_SETTINGS)
     @click.argument('LEVEL')
     @click.argument('MESSAGE')
     @click.argument('OBJECTS', nargs=-1)
     @click.option(
-        '--pretty/--ugly',
-        is_flag=True,
-        default=True,
-        help='Output JSON instead of key=value pairs for console logger')
+        '--pretty/--ugly', is_flag=True, default=True, help='Output JSON instead of key=value pairs for console logger'
+    )
     @click.option(
         '-j',
         '--json',
         'jsonify',
         is_flag=True,
         default=False,
-        help='Use the JSON logger formatter instead of the console one')
-    @click.option(
-        '-n',
-        '--name',
-        type=click.STRING,
-        default='Wryte',
-        help="Change the default logger's name")
-    @click.option(
-        '--no-color',
-        is_flag=True,
-        default=False,
-        help='Disable coloring in console formatter')
-    @click.option(
-        '--simple',
-        is_flag=True,
-        default=False,
-        help='Log only message to the console')
+        help='Use the JSON logger formatter instead of the console one',
+    )
+    @click.option('-n', '--name', type=click.STRING, default='Wryte', help="Change the default logger's name")
+    @click.option('--no-color', is_flag=True, default=False, help='Disable coloring in console formatter')
+    @click.option('--simple', is_flag=True, default=False, help='Log only message to the console')
     def main(level, message, objects, pretty, jsonify, name, no_color, simple):
-        wryter = Wryte(
-            name=name,
-            pretty=pretty,
-            level=level,
-            jsonify=jsonify,
-            color=not no_color,
-            simple=simple)
-        getattr(wryter, level.lower())(message, *objects)
+        wryter = Wryte(name=name, pretty=pretty, level=level, jsonify=jsonify, color=not no_color, simple=simple)
+
+        objcts = []
+
+        # Allows to pass `k=v` pairs.
+        for obj in objects:
+            try:
+                json.loads(obj)
+                objcts.append(obj)
+            except Exception:  # pylint: disable=broad-except
+                if '=' in obj:
+                    objcts.append(_split_kv(obj))
+
+        getattr(wryter, level.lower())(message, *objcts)
+
+
 else:
+
     def main():
         sys.exit(
             "To use Wryte's CLI you must first install certain dependencies. "
-            "Please run `pip install wryte[cli]` to enable the CLI.")
+            "Please run `pip install wryte[cli]` to enable the CLI."
+        )
